﻿import mapLoop from "./mapLoop";
import { Loop } from "redux-loop";
import { DataFilterMapReducer, DataFilterMapOptions, DataFilterAction, Filters, Options, DataFilterMapState, EntityCache, EntityGraph } from "./models";
import createApiCaller from "./createApiCaller";
import { defaultSpecs, defaultFilterState } from "./defaults";
import createFilterReducer from "./createFilterReducer";
import mergeEntities from "./mergeEntities";

export const initState = {
    entityCache: {},
    mounted: {},
    options: {}
};

export default <TFilterMap extends { [k in keyof TFilterMap]: any }, TDomain extends { [k in keyof TDomain]: any }>
    ({ apiMap }: DataFilterMapOptions<TFilterMap, TDomain>): DataFilterMapReducer<TFilterMap, TDomain> => {

    type S = DataFilterMapState<TFilterMap, TDomain>;
    type A = DataFilterAction<TFilterMap,TDomain>;
    type SLoop = S | Loop<S, A>;
    

    const updateEntityCache =
        (state: EntityCache<TFilterMap, TDomain>, updates: EntityGraph<TDomain>)
            : EntityCache<TFilterMap, TDomain> => {
            //console.log(updates);
            let newState: EntityCache<TFilterMap, TDomain> = {};
            const updatedEntityNames = Object.keys(updates) as (keyof TDomain)[];
            for (let i = 0, l = updatedEntityNames.length; i < l; ++i) {
                const name = updatedEntityNames[i];
                const oldSnapshotMap = state[name] || {} as any;
                newState[name] = mergeEntities(oldSnapshotMap, updates[name]!);
            }

            return {
                ...(state as any),
                ...(newState as any)
            };
    };
    
    // REDUCER

    return (state = initState, action) => {

        const { filterName } = action;

        

        if (action.type == "FILTER/MOUNT") {
            // merge options
            const options = {
                searchApi: apiMap[filterName],
                initialQuery: undefined,
                ...action.options
            };
            
            // initialize filter state
            const callApi = createApiCaller<TFilterMap, TDomain>(filterName);
            const filterState = options.initialQuery
                ? callApi(defaultFilterState, {
                    ...defaultSpecs,
                    ...options.initialQuery
                }, options)
                : defaultFilterState;

            // assemble global state
            return mapLoop(filterState, subState => ({
                ...state,
                options: {
                    ...(state.options as any),
                    [filterName]: options
                },
                mounted: {
                    ...(state.mounted as any),
                    [filterName]: subState
                }
            })) as SLoop;
        }
        else if (action.type == "FILTER/UNMOUNT") {
            const { [filterName]: oldOptions, ...restOfOptions } = state.options as any;
            const { [filterName]: oldFilterState, ...restOfMounted } = state.mounted as any;

            // assemble global state
            return {
                ...state,
                mounted: (restOfMounted as Filters<TFilterMap>),
                options: (restOfOptions as Options<TFilterMap, TDomain>)
            };
        }

        // ignore filter action if there is no mounted filter
        const filterBefore = state.mounted[filterName];
        if (filterBefore === undefined) return state;
        
        // reduce filter
        const updateFilterState = createFilterReducer(state.options[filterName]!);
        const filterAfter = updateFilterState(filterBefore!, action);

        // update entity cache from success API response
        const entityCache = action.type == "FILTER/REQUEST_SUCCESS"
            ? updateEntityCache(state.entityCache, action.payload.entities)
            : state.entityCache;
        
        // assemble global state
        return mapLoop(filterAfter, subState => ({
            ...state,
            entityCache,
            mounted: {
                ...(state.mounted as any),
                [filterName]: subState
            }
        })) as SLoop;
    };
}