﻿import { Loop } from "redux-loop";

export type DataStateSelector<TState, TFilterMap extends object, TDomain extends object> =
    (globalState: TState) => DataFilterMapState<TFilterMap, TDomain>;

export interface DataSpecs<TInput> {
    input: Partial<TInput>
    sortBy: string | null
    sortByDescending: boolean
    pageSize?: number
    pageIndex: number
}

export type EntityGraph<TDomain, TMetadata = {}> = {
    [K in keyof TDomain]?: {
        [id: string]: TDomain[K] & {
            __metadata?: TMetadata
        } 
    }
};

export interface DataFilterResponse<TDomain> {
    total: number
    matchKeys: string[]
    rootEntity?: keyof TDomain
    entities: EntityGraph<TDomain>
}

export interface EntityRef<TDomain, TK extends keyof TDomain> {
    __type: TK
    key: string
}

// USAGE
//
// interface ClassRoom {
//    id: string
//    name: string
//    capacity: number
//    attendees: Attendee[]
//    createdOn: Date
//    isActive: boolean
//}
//
//
//interface Attendee {
//    id: string
//    firstName: string
//    lastName: string
//    classrooms: ClassRoom[]
//}

//interface Entities {
//    classroom: ClassRoom,
//    attendee: Attendee
//}

//const response: DataFilterResponse<Entities> = {
//    total: 7,
//    matchKeys: ["1", "2"],
//    entities: {
//        attendee: {
//            "1": {
//                id: "1",
//                firstName: "Ahmad"
//            }
//        }
//    }
//};




export interface DataFilterState<TInput> {
    lastEntered: DataSpecs<TInput> | null
    lastApplied: DataSpecs<TInput> | null
    lastRequest: DataSpecs<TInput> | null
    isLoading: boolean
    total: number
    matchKeys: string[]
    error: any
}

export type DataFilterContainerActions<TInput> = {
    setPageIndex: (index: number) => void
    setPageSize: (size?: number) => void
    refresh: () => void
    setInput: (changes: Partial<TInput>) => void
    setSortBy: (path: string, descending?: boolean) => void
}

export interface DataFilterApiActions<TInput, TDomain>
    extends DataFilterContainerActions<TInput> {
    requestSuccess: (rs: DataFilterResponse<TDomain>) => void
    requestFailure: (err: any) => void
}

//type RefFactory<TDomain> = (type: keyof TDomain, key: string) => EntityRef<

export type SearchApi<TInput, TDomain> =
    (rq: DataSpecs<TInput>) => 
        Promise<DataFilterResponse<TDomain>>;

type PageIndexChangedAction<TFilterMap> = {
    type: "FILTER/PAGE_INDEX_CHANGED"
    filterName: keyof TFilterMap
    index: number
}

type PageSizeChangedAction<TFilterMap> = {
    type: "FILTER/PAGE_SIZE_CHANGED"
    filterName: keyof TFilterMap
    size?: number
}

type RefreshAction<TFilterMap> = {
    type: "FILTER/REFRESH"
    filterName: keyof TFilterMap
}

type FilterInputChangedAction<TFilterMap, TK extends keyof TFilterMap = keyof TFilterMap> = {
    type: "FILTER/INPUT_CHANGED"
    filterName: TK
    changes: Partial<TFilterMap[TK]>
}

type SortByChangedAction<TFilterMap> = {
    type: "FILTER/SORT_BY_CHANGED"
    filterName: keyof TFilterMap
    path: string
    descending?: boolean
}

export type RequestSuccessAction<TFilterMap, TDomain, TK extends keyof TFilterMap = keyof TFilterMap> = {
    type: "FILTER/REQUEST_SUCCESS"
    filterName: TK
    request: DataSpecs<TFilterMap[TK]>
    payload: DataFilterResponse<TDomain>
}



export type RequestErrorAction<TFilterMap, TK extends keyof TFilterMap = keyof TFilterMap> = {
    type: "FILTER/REQUEST_ERROR"
    filterName: TK
    request: DataSpecs<TFilterMap[TK]>
    payload: any
}

export type DataFilterMountAction<TFilterMap, TDomain, TK extends keyof TFilterMap = keyof TFilterMap> = {
    type: "FILTER/MOUNT"
    filterName: TK
    options: Partial<DataFilterOptions<TFilterMap[TK], TDomain>>
}

export type DataFilterUnmountAction<TFilterMap, TK extends keyof TFilterMap = keyof TFilterMap> = {
    type: "FILTER/UNMOUNT"
    filterName: TK
}

export type DataFilterAction<TFilterMap, TDomain> =
    PageIndexChangedAction<TFilterMap> |
    PageSizeChangedAction<TFilterMap> |
    RefreshAction<TFilterMap> |
    FilterInputChangedAction<TFilterMap> |
    SortByChangedAction<TFilterMap> |
    RequestSuccessAction<TFilterMap, TDomain> |
    RequestErrorAction<TFilterMap> |
    DataFilterMountAction<TFilterMap, TDomain> |
    DataFilterUnmountAction<TFilterMap>;
    
export type DataFilterOptions<TInput, TDomain> = {
    //filterName: string
    searchApi: SearchApi<TInput,TDomain> | undefined
    initialQuery: Partial<DataSpecs<TInput>> | undefined
}

export type Filters<TFilterMap extends object> = {
    [N in keyof TFilterMap]?: DataFilterState<TFilterMap[N]>
};

export type Options<TFilterMap,TDomain> = {
    [N in keyof TFilterMap]?: DataFilterOptions<TFilterMap[N], TDomain> 
};

export type EntityCache<TFilterMap, TDomain> = EntityGraph<TDomain, {
    [referencedBy in keyof TFilterMap]?: boolean
}>

export type DataFilterMapState<TFilterMap extends object,TDomain extends object> = {
    entityCache: EntityCache<TFilterMap,TDomain>
    mounted: Filters<TFilterMap>
    options: Options<TFilterMap,TDomain>
}

export type DataFilterMapReducer<TFilterMap extends object, TDomain extends object> =
    (state: DataFilterMapState<TFilterMap, TDomain> | undefined, action: DataFilterAction<TFilterMap, TDomain>) =>
        DataFilterMapState<TFilterMap, TDomain> |
        Loop<DataFilterMapState<TFilterMap, TDomain>, DataFilterAction<TFilterMap, TDomain>>;

export type DataFilterMapOptions<TFilterMap, TDomain> = {
    apiMap: { [N in keyof TFilterMap]?: SearchApi<TFilterMap[N], TDomain> }
};
